"""Functions to batch process trades into dataframes for analysis.
"""
from __future__ import annotations
from typing import Tuple, List

from .live_odds import live_odds
from .qpbanker import win_probability, expected_value, average_odds
from .optimization import _pareto_filter

import polars as pl
import numpy as np
from itertools import combinations
from tqdm import tqdm


def _all_subsets(lst): return [list(x) for r in range(
    1, len(lst)+1) for x in combinations(lst, r)]  # list subsets of a list


def _process_single_qp_trade(banker: int, covered: List[int], odds_pla: List[float], odds_qpl: List[float], rebate: float) -> Tuple[int, List, float, float]:
    """Process a single qp trade.
    """
    win_prob = win_probability(odds_pla, banker, covered)
    exp_value = expected_value(odds_pla, odds_qpl, banker, covered, rebate)
    ave_odds = average_odds(odds_qpl, banker, covered)
    return (banker, covered, win_prob, exp_value, ave_odds)


def generate_all_qp_trades(date: str, venue_code: str, race_number: int, rebate: float = 0.12, fit_harville=False) -> pl.DataFrame:
    """Generate all possible qp tickets for the specified race.

    Args:
        date (str): Date in 'YYYY-MM-DD' format.
        venue_code (str): Venue code, e.g., 'ST' for Shatin, 'HV' for Happy Valley.
        race_number (int): Race number.
        rebate (float, optional): The rebate percentage. Defaults to 0.12.
        fit_harville (bool, optional): Whether to fit the odds using Harville model. Defaults to False.

    Returns:
        pl.DataFrame: DataFrame with all possible trades and their metrics.
    """

    odds = live_odds(date, venue_code, race_number,
                     odds_type=['PLA', 'QPL'], fit_harville=fit_harville)
    N = len(odds['PLA'])
    candidates = np.arange(1, N+1)

    results = [_process_single_qp_trade(banker, covered, odds['PLA'], odds['QPL'], rebate)
               for banker in tqdm(candidates, desc="Processing bankers")
               for covered in _all_subsets(candidates[candidates != banker])]

    df = (pl.DataFrame(results, schema=['Banker', 'Covered', 'WinProb', 'ExpValue', 'AvgOdds'])
          .with_columns(pl.col('Covered').list.len().alias('NumCovered')))

    return df


def generate_pareto_qp_trades(date: str, venue_code: str, race_number: int, rebate: float = 0.12, groupby: List[str] = [], fit_harville=False) -> pl.DataFrame:
    """Generate qp tickets that are Pareto optimal for the specified race.

    Args:
        date (str): Date in 'YYYY-MM-DD' format.
        venue_code (str): Venue code, e.g., 'ST' for Shatin, 'HV' for Happy Valley.
        race_number (int): Race number.
        rebate (float, optional): The rebate percentage. Defaults to 0.12.
        groupby (List[str], optional): Columns to group by when determining Pareto optimality. Defaults to [] (global optimal).
        harville_fit (bool, optional): Whether to fit the odds using Harville model. Defaults to False.

    Returns:
        pl.DataFrame: DataFrame with all Pareto trades and their metrics.
    """
    df = generate_all_qp_trades(date, venue_code, race_number, rebate, harville_fit=fit_harville)
    pareto_df = _pareto_filter(df, groupby=groupby, by=[
                               'WinProb', 'ExpValue'], maximize=True)
    return pareto_df



import polars as pl
from typing import List, Union

def _pareto_filter(
    df: pl.DataFrame, 
    groupby: List[str], 
    by: List[str], 
    maximize: Union[bool, List[bool]] = True
) -> pl.DataFrame:
    """
    Filter dataframe to only include Pareto optimal rows within each group.

    Args:
        df: Input dataframe
        groupby: Columns to group by (empty list for global filter)
        by: Columns to consider for Pareto optimality
        maximize: Whether to maximize (True) or minimize (False) each 'by' column
    
    Returns:
        DataFrame containing only Pareto optimal rows
    """
    if df.is_empty() or not by:
        return df

    # Normalize maximize to list
    maximize_list = [maximize] * len(by) if isinstance(maximize, bool) else maximize
    
    if len(maximize_list) != len(by):
        raise ValueError(f"Length of 'maximize' ({len(maximize_list)}) must equal length of 'by' ({len(by)})")

    # Single objective: simple min/max filter
    if len(by) == 1:
        opt_expr = pl.col(by[0]).max() if maximize_list[0] else pl.col(by[0]).min()
        if groupby:
            opt_expr = opt_expr.over(groupby)
        return df.filter(pl.col(by[0]) == opt_expr)

    # Two objectives: efficient skyline algorithm
    if len(by) == 2:
        temp_cols = ["__obj_0", "__obj_1"]
        
        # Transform to maximization problem
        df_temp = df.with_columns([
            (pl.col(by[i]) * (1 if maximize_list[i] else -1)).alias(temp_cols[i])
            for i in range(2)
        ])
        
        # Sort by first objective descending, then second descending (for stability)
        groupby = groupby or []
        sort_cols = (groupby if groupby else []) + temp_cols
        sorted_df = df_temp.sort(sort_cols, descending=[False] * len(groupby) + [True, True])
        
        # Keep rows where second objective is not dominated by any previous row in group
        if groupby:
            max_so_far = pl.col(temp_cols[1]).cum_max().shift(1, fill_value=float("-inf")).over(groupby)
        else:
            max_so_far = pl.col(temp_cols[1]).cum_max().shift(1, fill_value=float("-inf"))
        
        mask = pl.col(temp_cols[1]) > max_so_far
        return sorted_df.filter(mask).drop(temp_cols)

    # N objectives (N > 2): pairwise dominance check
    df_with_id = df.with_row_index("__id")
    
    # Self-join to compare all pairs
    left = df_with_id.lazy()
    right = df_with_id.lazy()
    
    if groupby:
        pairs = left.join(right, on=groupby, suffix="_r")
    else:
        pairs = left.join(right, how="cross", suffix="_r")
    
    # Only compare different rows
    pairs = pairs.filter(pl.col("__id") != pl.col("__id_r"))
    
    # Build dominance conditions
    dominance_conditions = []
    for col, is_max in zip(by, maximize_list):
        if is_max:
            # right dominates left if right[col] >= left[col] for all cols
            dominance_conditions.append(pl.col(f"{col}_r") >= pl.col(col))
        else:
            dominance_conditions.append(pl.col(f"{col}_r") <= pl.col(col))
    
    # Strict dominance: all >= and at least one >
    strict_conditions = []
    for col, is_max in zip(by, maximize_list):
        if is_max:
            strict_conditions.append(pl.col(f"{col}_r") > pl.col(col))
        else:
            strict_conditions.append(pl.col(f"{col}_r") < pl.col(col))
    
    is_dominated = pl.all_horizontal(dominance_conditions) & pl.any_horizontal(strict_conditions)
    
    # Find IDs of dominated rows
    dominated_ids = (
        pairs.filter(is_dominated)
        .select("__id")
        .unique()
        .collect()
        .get_column("__id")
    )
    
    # Return non-dominated rows
    return df_with_id.filter(~pl.col("__id").is_in(dominated_ids)).drop("__id")


"""Functions to perform probability and expectation calculations for the QP Banker strategy.
"""
from __future__ import annotations

from typing import List
import numpy as np

def _pla_odds_partition(pla_odds: np.ndarray, banker: int, covered: List[int]) -> tuple[float, float, float]:
    """Partition the place odds into banker, covered and eliminated sets.

    Args:
        pla_odds (np.ndarray): An array of place odds for the horses (0-indexed).
        banker (int): The horse number of the banker (1-indexed).
        covered (List[int]): A list of horse numbers in the cover set (1-indexed).

    Returns:
        tuple[float, float, float]: A tuple containing the probabilities of the banker,
                                    covered set and eliminated set.
    """
    p_banker = 1 / pla_odds[banker - 1]
    Z_covered = sum([1/pla_odds[c-1] for c in covered])
    Z_total = (1/pla_odds).sum()
    Z_elim = Z_total - p_banker - Z_covered

    return p_banker, Z_covered, Z_elim

def _double_win_probability(pla_odds: np.ndarray, banker: int, covered: List[int]) -> float:
    """Calculate the probability of winning a two tickets in the QPBanker strategy.
    See overleaf document for derivation.

    Args:
        pla_odds (np.ndarray): An array of place odds for the horses (0-indexed).
        banker (int): The horse number of the banker (1-indexed).
        covered (List[int]): A list of horse numbers in the cover set (1-indexed).

    Returns:
        float: probability
    """
    p_banker, Z_covered, Z_elim = _pla_odds_partition(pla_odds, banker, covered)
    Z_total = p_banker + Z_covered + Z_elim

    return 3*p_banker*Z_covered**2 / Z_total**3


def _single_win_probability(pla_odds: np.ndarray, banker: int, covered: List[int]) -> float:
    """Calculate the probability of winning a single ticket in the QPBanker strategy.
    See overleaf document for derivation.

    Args:
        pla_odds (np.ndarray): An array of place odds for the horses (0-indexed).
        banker (int): The horse number of the banker (1-indexed).
        covered (List[int]): A list of horse numbers in the cover set (1-indexed).

    Returns:
        float: probability
    """
    p_banker, Z_covered, Z_elim = _pla_odds_partition(pla_odds, banker, covered)
    Z_total = p_banker + Z_covered + Z_elim

    return 6*p_banker*Z_covered*Z_elim / Z_total**3


def win_probability(pla_odds: np.ndarray, banker: int, covered: List[int]) -> float:
    """Calculate the probability of winning at least one ticket in the QPBanker strategy.

    Args:
        pla_odds (np.ndarray): An array of place odds for the horses (0-indexed).
        banker (int): The horse number of the banker (1-indexed).
        covered (List[int]): A list of horse numbers in the cover set (1-indexed).

    Returns:
        float: probability
    """
    p_double = _double_win_probability(pla_odds, banker, covered)
    p_single = _single_win_probability(pla_odds, banker, covered)

    return p_double + p_single


def expected_value(pla_odds: np.ndarray, qpl_odds: np.ndarray, banker: int, covered: List[int], rebate: float = 0.12) -> float:
    """Calculate the expected value (per dollar) of the QPBanker strategy using constant stake.

    Args:
        pla_odds (np.ndarray): An array of place odds for the horses (0-indexed).
        qpl_odds (np.ndarray): An array of quinella place odds for the horses (0-indexed).
        banker (int): The horse number of the banker (1-indexed).
        covered (List[int]): A list of horse numbers in the cover set (1-indexed).
        rebate (float, optional): The rebate percentage. Defaults to 0.12.
    
    Returns:
            float: expected value per dollar staked
    """
    p_banker, Z_covered, Z_elim = _pla_odds_partition(pla_odds, banker, covered)
    Z_total = p_banker + Z_covered + Z_elim
    pla_prob = 1/pla_odds

    P_dbl = _double_win_probability(pla_odds, banker, covered)
    P_single = _single_win_probability(pla_odds, banker, covered)
    C = len(covered)

    BigSum = sum([qpl_odds[banker-1][c-1]*pla_prob[c-1]*(Z_covered-pla_prob[c-1]+Z_elim) for c in covered])
    EV = 6*p_banker*BigSum / Z_total**3 - (1-rebate)*C - rebate*(2*P_dbl + P_single)

    return EV / C

def average_odds(qpl_odds: np.ndarray, banker: int, covered: List[int]) -> float:
    """Calculate the (harmonic) average odds across the covered set.

    Args:
        qpl_odds (np.ndarray): An array of quinella place odds for the horses (0-indexed).
        banker (int): The horse number of the banker (1-indexed).
        covered (List[int]): A list of horse numbers in the cover set (1-indexed).
        
    Returns:
            float: average odds
    """
    C = len(covered)
    avg_odds = C / sum([1/qpl_odds[banker-1][c-1] for c in covered])
    return avg_odds


"""Functions to fetch and process historical race and horse data from HKJC
"""
from __future__ import annotations

import requests

# TODO read and process all races from start date to end date

# TODO query all basic info and race history for a specific horse

# TODO classify running style & draw to determine blocking probability

# 


"""
Harville Race Model Optimizer

Estimates horse racing outcome probabilities using the Harville model via dynamic
programming. Fits latent strength parameters from observed betting market odds across
multiple pool types (Win, Qin, Quinella, Banker).

The optimizer uses O(N * 2^N) complexity DP with Numba JIT compilation for speed.
Suitable for races with up to ~20 horses.

Example:
    >>> optimizer = HarvilleOptimizer(n_horses=14)
    >>> results = optimizer.fit(W_obs=win_odds, Qin_obs=qin_odds, 
    ...                          Q_obs=quinella_odds, b_obs=banker_odds)
    >>> print(results['theta'])  # Fitted strength parameters
"""

import numpy as np
from scipy.optimize import minimize
from numba import njit
from typing import Tuple, Optional

LAMBDA_DEFAULTS = {
    'WIN': 1.0,
    'QIN': 2.0,
    'QPL': 1.5,
    'PLA': 0.7
}

@njit(cache=True)
def _popcount(mask: int) -> int:
    count = 0
    while mask:
        count += 1
        mask &= mask - 1
    return count


@njit(cache=True)
def _precompute_mask_info(n: int) -> Tuple[np.ndarray, np.ndarray]:
    max_mask = 1 << n
    mask_strength_coef = np.zeros((max_mask, n), dtype=np.float64)
    mask_popcount = np.zeros(max_mask, dtype=np.int32)
    
    for mask in range(max_mask):
        mask_popcount[mask] = _popcount(mask)
        for i in range(n):
            if mask & (1 << i):
                mask_strength_coef[mask, i] = 1.0
    
    return mask_strength_coef, mask_popcount


@njit(cache=True)
def _compute_dp_vectorized(theta: np.ndarray, k_max: int) -> np.ndarray:
    n = len(theta)
    max_mask = 1 << n
    
    mask_strength_coef, mask_popcount = _precompute_mask_info(n)
    mask_strength = mask_strength_coef @ theta
    
    dp = np.zeros((k_max + 1, max_mask))
    dp[0, 0] = 1.0
    
    for k in range(k_max):
        valid_masks = np.where(mask_popcount == k)[0]
        
        for mask in valid_masks:
            if dp[k, mask] == 0:
                continue
            
            s_mask = mask_strength[mask]
            remaining = 1.0 - s_mask
            
            if remaining < 1e-12:
                continue
            
            prob_current = dp[k, mask]
            
            for i in range(n):
                if not (mask & (1 << i)):
                    next_mask = mask | (1 << i)
                    dp[k + 1, next_mask] += prob_current * theta[i] / remaining
    
    return dp


@njit(cache=True)
def _extract_pair_in_top_k(dp: np.ndarray, n: int, k: int) -> np.ndarray:
    M = np.zeros((n, n))
    max_mask = 1 << n
    
    mask_popcount = np.zeros(max_mask, dtype=np.int32)
    for mask in range(max_mask):
        mask_popcount[mask] = _popcount(mask)
    
    masks_size_k = np.where(mask_popcount == k)[0]
    
    for mask in masks_size_k:
        prob = dp[k, mask]
        if prob == 0:
            continue
        
        horses = np.empty(k, dtype=np.int32)
        idx = 0
        for i in range(n):
            if mask & (1 << i):
                horses[idx] = i
                idx += 1
        
        for i in range(k):
            for j in range(k):
                if horses[i] != horses[j]:
                    M[horses[i], horses[j]] += prob
    
    return M


@njit(cache=True)
def _extract_top_k_probs(dp: np.ndarray, n: int, k_max: int) -> np.ndarray:
    T = np.zeros((n, k_max + 1))
    max_mask = 1 << n
    
    mask_popcount = np.zeros(max_mask, dtype=np.int32)
    for mask in range(max_mask):
        mask_popcount[mask] = _popcount(mask)
    
    for k in range(1, k_max + 1):
        masks_size_k = np.where(mask_popcount == k)[0]
        
        for mask in masks_size_k:
            prob = dp[k, mask]
            if prob == 0:
                continue
            
            for i in range(n):
                if mask & (1 << i):
                    T[i, k] += prob
    
    return T


@njit(cache=True)
def _compute_probabilities(theta: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    n = len(theta)
    
    dp = _compute_dp_vectorized(theta, n)
    
    T = _extract_top_k_probs(dp, n, n)
    
    P = np.zeros((n, n))
    for i in range(n):
        for j in range(n):
            P[i, j] = T[i, j + 1] - T[i, j]
    
    W = P[:, 0]
    Qin = _extract_pair_in_top_k(dp, n, 2)
    Q = _extract_pair_in_top_k(dp, n, 3)
    b = T[:, 3]
    
    return W, Qin, Q, b, P


@njit(cache=True)
def _kl_divergence(p_obs: np.ndarray, p_model: np.ndarray) -> float:
    eps = 1e-10
    
    p_obs_flat = np.maximum(p_obs.ravel(), eps)
    p_model_flat = np.maximum(p_model.ravel(), eps)
    
    sum_obs = p_obs_flat.sum()
    sum_model = p_model_flat.sum()
    
    if sum_obs > eps:
        p_obs_flat = p_obs_flat / sum_obs
    if sum_model > eps:
        p_model_flat = p_model_flat / sum_model
    
    return np.sum(p_obs_flat * np.log(p_obs_flat / p_model_flat))


class HarvilleModel:
    """
    Fits Harville race model to betting market odds using dynamic programming.
    
    The Harville model assigns each horse a latent strength parameter theta_i, where
    the probability of finishing next among remaining horses is proportional to 
    relative strength. This optimizer estimates theta from observed betting odds
    across multiple pool types.
    
    Default lambda weights reflect that early Win odds are biased by informed 
    traders waiting until closing, while exotic pools provide more stable 
    signals for ensemble estimation.
    
    Attributes:
        n (int): Number of horses
        lambda_win (float): Weight for Win pool loss
        lambda_qin (float): Weight for Qin pool loss
        lambda_quinella (float): Weight for Quinella pool loss
        lambda_banker (float): Weight for Banker pool loss
    """
    
    def __init__(self, n_horses: int, lambda_win: float = LAMBDA_DEFAULTS['WIN'], lambda_qin: float = LAMBDA_DEFAULTS['QIN'], 
                 lambda_quinella: float = LAMBDA_DEFAULTS['QPL'], lambda_banker: float = LAMBDA_DEFAULTS['PLA']) -> None:
        """
        Initialize model.
        
        Args:
            n_horses: Number of horses in race (recommend <= 20 for speed)
            lambda_win: Weight for Win odds (prob horse finishes 1st)
            lambda_qin: Weight for Qin odds (prob pair finishes 1st-2nd)
            lambda_quinella: Weight for Quinella odds (prob pair in top 3)
            lambda_banker: Weight for Banker odds (prob horse in top 3)
            
        Raises:
            ValueError: If n_horses > 20 (exponential complexity warning)
        """
        if n_horses > 20:
            raise ValueError("N > 20 may be too slow (2^N complexity)")
        
        self.n = n_horses
        self.lambda_win = lambda_win
        self.lambda_qin = lambda_qin
        self.lambda_quinella = lambda_quinella
        self.lambda_banker = lambda_banker
        self._eval_count = 0
        self.result = None
    
    def _loss(self, theta: np.ndarray, W_obs: Optional[np.ndarray], 
             Qin_obs: Optional[np.ndarray], Q_obs: Optional[np.ndarray], 
             b_obs: Optional[np.ndarray]) -> float:
        """
        Compute weighted KL divergence loss between observed and model odds.
        
        Args:
            theta: Strength parameters (will be normalized to simplex)
            W_obs: Observed Win probabilities (n,) or None
            Qin_obs: Observed Qin probabilities (n, n) or None
            Q_obs: Observed Quinella probabilities (n, n) or None
            b_obs: Observed Banker probabilities (n,) or None
            
        Returns:
            Scalar loss value (sum of weighted KL divergences)
        """
        self._eval_count += 1
        
        theta = np.abs(theta) + 1e-10
        theta = theta / theta.sum()
        
        W_model, Qin_model, Q_model, b_model, P_model = _compute_probabilities(theta)
        
        loss = 0.0
        
        if W_obs is not None:
            loss += self.lambda_win * _kl_divergence(W_obs, W_model)
        
        if Qin_obs is not None:
            loss += self.lambda_qin * _kl_divergence(Qin_obs, Qin_model)
        
        if Q_obs is not None:
            loss += self.lambda_quinella * _kl_divergence(Q_obs, Q_model)
        
        if b_obs is not None:
            loss += self.lambda_banker * _kl_divergence(b_obs, b_model)
        
        return loss
    
    def fit(self, W_obs: Optional[np.ndarray] = None, 
            Qin_obs: Optional[np.ndarray] = None,
            Q_obs: Optional[np.ndarray] = None, 
            b_obs: Optional[np.ndarray] = None,
            theta_init: Optional[np.ndarray] = None, 
            method: str = 'L-BFGS-B') -> dict:
        """
        Fit Harville model to observed betting odds.
        
        At least one odds type must be provided. All odds should be probabilities
        (not decimal/fractional odds). Matrices should be symmetric where applicable.
        
        Args:
            W_obs: Win probabilities, shape (n,). W_obs[i] = prob horse i wins
            Qin_obs: Qin probabilities, shape (n, n). Qin_obs[i,j] = prob horses
                     i,j finish 1st-2nd in any order
            Q_obs: Quinella probabilities, shape (n, n). Q_obs[i,j] = prob horses
                   i,j both finish in top 3
            b_obs: Banker probabilities, shape (n,). b_obs[i] = prob horse i 
                   finishes in top 3
            theta_init: Initial strength guess (default: W_obs if available, else uniform)
            method: Scipy optimizer ('L-BFGS-B' or 'SLSQP')
            
        Returns:
            Dictionary containing:
                - theta: Fitted strength parameters (n,)
                - W_fitted: Fitted Win probabilities (n,)
                - Qin_fitted: Fitted Qin probabilities (n, n)
                - Q_fitted: Fitted Quinella probabilities (n, n)
                - b_fitted: Fitted Banker probabilities (n,)
                - P_fitted: Full place probability matrix (n, n), P[i,j] = 
                            prob horse i finishes in position j
                - loss: Final loss value
                - success: Whether optimization converged
                - message: Optimizer status message
                - n_eval: Number of loss function evaluations
                
        Raises:
            ValueError: If no odds provided or shapes don't match n_horses
            
        Example:
            >>> opt = HarvilleOptimizer(n_horses=10)
            >>> results = opt.fit(W_obs=win_probs, Q_obs=quinella_probs)
            >>> print(f"Fitted strengths: {results['theta']}")
            >>> print(f"Converged: {results['success']}")
        """
        if W_obs is None and Qin_obs is None and Q_obs is None and b_obs is None:
            raise ValueError("At least one type of odds must be provided")
        
        if W_obs is not None and W_obs.shape != (self.n,):
            raise ValueError(f"W_obs must be ({self.n},)")
        if Qin_obs is not None and Qin_obs.shape != (self.n, self.n):
            raise ValueError(f"Qin_obs must be ({self.n}, {self.n})")
        if Q_obs is not None and Q_obs.shape != (self.n, self.n):
            raise ValueError(f"Q_obs must be ({self.n}, {self.n})")
        if b_obs is not None and b_obs.shape != (self.n,):
            raise ValueError(f"b_obs must be ({self.n},)")
        
        if theta_init is None:
            if W_obs is not None:
                theta_init = W_obs / W_obs.sum()
            else:
                theta_init = np.ones(self.n) / self.n
        else:
            theta_init = theta_init / theta_init.sum()
        
        self._eval_count = 0
        
        if method == 'L-BFGS-B':
            result = minimize(
                fun=lambda x: self._loss(x, W_obs, Qin_obs, Q_obs, b_obs),
                x0=theta_init,
                method='L-BFGS-B',
                bounds=[(1e-6, 1.0) for _ in range(self.n)],
                options={'maxiter': 500, 'ftol': 1e-9, 'maxls': 50}
            )
        else:
            result = minimize(
                fun=lambda x: self._loss(x, W_obs, Qin_obs, Q_obs, b_obs),
                x0=theta_init,
                method='SLSQP',
                bounds=[(1e-6, 1.0) for _ in range(self.n)],
                constraints={'type': 'eq', 'fun': lambda x: x.sum() - 1},
                options={'maxiter': 500, 'ftol': 1e-9}
            )
        
        theta_opt = np.abs(result.x) + 1e-10
        theta_opt = theta_opt / theta_opt.sum()
        
        W_fitted, Qin_fitted, Q_fitted, b_fitted, P_fitted = _compute_probabilities(theta_opt)
        
        self.result = {
            'theta': theta_opt,
            'W_fitted': W_fitted,
            'Qin_fitted': Qin_fitted,
            'Q_fitted': Q_fitted,
            'b_fitted': b_fitted,
            'P_fitted': P_fitted,
            'loss': result.fun,
            'success': result.success,
            'message': result.message,
            'n_eval': self._eval_count
        }

        return self.result
    
def fit_harville_to_odds(odds : dict[str, np.ndarray], lambdas : dict[str, float] = None) -> dict:
    """
    Fit Harville model to observed betting odds.
    
    At least one odds type must be provided. All odds should be probabilities
    (not decimal/fractional odds). Matrices should be symmetric where applicable.
    
    Args:
        odds: Dictionary of odds arrays with types as keys.:
                'WIN' (n,), 'QIN' (n,n), 'QPL' (n,n), 'PLA' (n,)
        lambdas: Optional dictionary of lambda weights for each odds type.
                    Keys can be 'WIN', 'QIN', 'QPL', 'PLA'. Defaults to
                    {'WIN': 1.0, 'QIN': 2.0, 'QPL': 1.5, 'PLA': 0.7}
        
    Returns:
        Dictionary containing:
            - theta: Fitted strength parameters (n,)
            - W_fitted: Fitted Win probabilities (n,)
            - Qin_fitted: Fitted Qin probabilities (n, n)
            - Q_fitted: Fitted Quinella probabilities (n, n)
            - b_fitted: Fitted Banker probabilities (n,)
            - P_fitted: Full place probability matrix (n, n), P[i,j] = 
                        prob horse i finishes in position j
            - loss: Final loss value
            - success: Whether optimization converged
            - message: Optimizer status message
            - n_eval: Number of loss function evaluations
    """
    n_horses = None
    W_obs = None
    Qin_obs = None
    Q_obs = None
    b_obs = None
    
    if 'WIN' in odds:
        W_odds = odds['WIN']
        if n_horses is None:
            n_horses = len(W_odds)
        elif n_horses != len(W_odds):
            raise ValueError("Inconsistent number of horses in WIN odds")
        W_obs = np.nan_to_num(1.0 / W_odds, 0)
    
    if 'QIN' in odds:
        Qin_odds = odds['QIN']
        if n_horses is None:
            n_horses = Qin_odds.shape[0]
        elif n_horses != Qin_odds.shape[0]:
            raise ValueError("Inconsistent number of horses in QIN odds")
        Qin_obs = np.nan_to_num(1.0 / Qin_odds, 0)
    
    if 'QPL' in odds:
        Q_odds = odds['QPL']
        if n_horses is None:
            n_horses = Q_odds.shape[0]
        elif n_horses != Q_odds.shape[0]:
            raise ValueError("Inconsistent number of horses in QPL odds")
        Q_obs = np.nan_to_num(1.0 / Q_odds, 0)
    
    if 'PLA' in odds:
        b_odds = odds['PLA']
        if n_horses is None:
            n_horses = len(b_odds)
        elif n_horses != len(b_odds):
            raise ValueError("Inconsistent number of horses in PLA odds")
        b_obs = np.nan_to_num(1.0 / b_odds, 0)
    
    merged_lambdas = {**LAMBDA_DEFAULTS, **(lambdas or {})}
    ho = HarvilleModel(
        n_horses,
        lambda_win=merged_lambdas['WIN'],
        lambda_qin=merged_lambdas['QIN'],
        lambda_quinella=merged_lambdas['QPL'],
        lambda_banker=merged_lambdas['PLA']
    )
    result = ho.fit(W_obs=W_obs, Qin_obs=Qin_obs, Q_obs=Q_obs, b_obs=b_obs)
    return result


"""Functions to fetch and process data from HKJC
"""
from __future__ import annotations
from typing import Tuple, List

from .harville_model import fit_harville_to_odds

import requests
from cachetools.func import ttl_cache
import numpy as np
from datetime import datetime as dt

ENDPOINT = "https://info.cld.hkjc.com/graphql/base/"

LIVEODDS_PAYLOAD = {
    "operationName": "racing",
    "variables": {"date": None, "venueCode": None, "raceNo": None, "oddsTypes": None},
    "query": """
query racing($date: String, $venueCode: String, $oddsTypes: [OddsType], $raceNo: Int) {
    raceMeetings(date: $date, venueCode: $venueCode) {
        pmPools(oddsTypes: $oddsTypes, raceNo: $raceNo) {
            id
            status
            sellStatus
            oddsType
            lastUpdateTime
            guarantee
            minTicketCost
            name_en
            name_ch
            leg {
                number
                races
            }
            cWinSelections {
                composite
                name_ch
                name_en
                starters
            }
            oddsNodes {
                combString
                oddsValue
                hotFavourite
                oddsDropValue
                bankerOdds {
                    combString
                    oddsValue
                }
            }
        }
    }
}""",
}


@ttl_cache(maxsize=12, ttl=30)
def _fetch_live_odds(date: str, venue_code: str, race_number: int, odds_type: Tuple[str] = ('PLA', 'QPL')) -> Tuple[dict]:
    """Fetch live odds data from HKJC GraphQL endpoint."""
    payload = LIVEODDS_PAYLOAD.copy()
    payload["variables"] = payload["variables"].copy()
    payload["variables"]["date"] = date
    payload["variables"]["venueCode"] = venue_code
    payload["variables"]["raceNo"] = race_number
    payload["variables"]["oddsTypes"] = odds_type

    headers = {
        "Origin": "https://bet.hkjc.com",
        "Referer": "https://bet.hkjc.com",
        "Content-Type": "application/json",
        "Accept": "application/json",
        "User-Agent": "python-hkjc-fetch/0.1",
    }

    r = requests.post(ENDPOINT, json=payload, headers=headers, timeout=10)
    if r.status_code != 200:
        raise RuntimeError(f"Request failed: {r.status_code} - {r.text}")

    meetings = r.json().get("data", {}).get("raceMeetings", [])

    return [
        {"HorseID": node["combString"], "Type": pool.get(
            "oddsType"), "Odds": float(node["oddsValue"])}
        for meeting in meetings
        for pool in meeting.get("pmPools", [])
        for node in pool.get("oddsNodes", [])
    ]


def live_odds(date: str, venue_code: str, race_number: int, odds_type: List[str] = ['PLA', 'QPL'], fit_harville=False) -> dict:
    """Fetch live odds as numpy arrays.

    Args:
        date (str): Date in 'YYYY-MM-DD' format.
        venue_code (str): Venue code, e.g., 'ST' for Shatin, 'HV' for Happy Valley.
        race_number (int): Race number.
        odds_type (List[str]): Types of odds to fetch. Default is ['PLA', 'QPL']. Currently the following types are supported:
            - 'WIN': Win odds
            - 'PLA': Place odds
            - 'QIN': Quinella odds
            - 'QPL': Quinella Place odds
        fit_harville (bool): Whether to fit the odds using Harville model. Default is False.

    Returns:
        dict: Dictionary with keys as odds types and values as numpy arrays containing the odds.
            If odds_type is 'WIN','PLA', returns a 1D array of place odds.
            If odds_type is 'QIN','QPL', returns a 2D array of quinella place odds. 
    """
    # validate date format
    try:
        dt.strptime(date, "%Y-%m-%d")
    except Exception:
        raise ValueError("Date must be in 'YYYY-MM-DD' format")

    mandatory_types = ('WIN','PLA','QIN','QPL') if fit_harville else ('PLA')

    data = _fetch_live_odds(date, venue_code, race_number,
                            odds_type=tuple(set(mandatory_types+odds_type)))

    # use place odds to determine number of horses
    pla_data = [entry for entry in data if entry["Type"] == "PLA"]
    N = len(pla_data)

    odds = {'WIN': np.full(N, np.nan, dtype=float),
            'PLA': np.full(N, np.nan, dtype=float),
            'QIN': np.full((N, N), np.nan, dtype=float),
            'QPL': np.full((N, N), np.nan, dtype=float)}

    for entry in data:
        if entry["Type"] in ["QIN", "QPL"]:
            horse_ids = list(map(int, entry["HorseID"].split(",")))
            odds[entry["Type"]][horse_ids[0] - 1, horse_ids[1] - 1] = entry["Odds"]
            odds[entry["Type"]][horse_ids[1] - 1, horse_ids[0] - 1] = entry["Odds"]
        elif entry["Type"] in ["PLA","WIN"]:
            odds[entry["Type"]][int(entry["HorseID"]) - 1] = entry["Odds"]

    if fit_harville:
        fit_res = fit_harville_to_odds(
            W_obs=odds['WIN'],
            Qin_obs=odds['QIN'],
            Q_obs=odds['QPL'],
            b_obs=odds['PLA']
        )
        if fit_res['success']:
            odds['PLA'] = np.nan_to_num(1/fit_res['b_fitted'], posinf=0)
            odds['QPL'] = np.nan_to_num(1/fit_res['Q_fitted'], posinf=0)
            odds['WIN'] = np.nan_to_num(1/fit_res['W_fitted'], posinf=0)
            odds['QIN'] = np.nan_to_num(1/fit_res['Qin_fitted'], posinf=0)
        else:
            print(f"[WARNING] Harville model fitting failed: {fit_res.get('message','')}")

    return {t: odds[t] for t in odds_type}



"""Top-level package for hkjc tools.

This module re-exports commonly used symbols from the submodules.
"""
from importlib.metadata import version as _version

__all__ = ["live_odds", "qpbanker",
		    "generate_all_qp_trades", "generate_pareto_qp_trades",
			"speedpro_energy", "speedmap","harveille_model"]

try:
	__version__ = _version(__name__)
except Exception:  # pragma: no cover - best-effort version resolution
	__version__ = "0.0.0"

from .live_odds import live_odds
from .processing import generate_all_qp_trades, generate_pareto_qp_trades
from .speedpro import speedmap, speedpro_energy
from . import harville_model


"""Functions to scrap SpeedPro data.
"""
from __future__ import annotations

from datetime import datetime as dt

import polars as pl
import requests
import json

ENERGY_XLS_TEMPLATE = "https://racing.hkjc.com/racing/speedpro/assets/excel/{date}_energygrid_en.xls"
SPEEDMAP_URL_TEMPLATE = "https://racing.hkjc.com/racing/speedpro/assets/json/speedguide/race_{race_num}.json"


def speedpro_energy(race_date: str) -> pl.DataFrame:
    """Fetch and process SpeedPro scores for a given race date.

    Args:
        race_date (str): Date in 'YYYY-MM-DD' format.

    Returns:
        pl.DataFrame: Processed DataFrame with SpeedPro scores.
    """
    # validate date format
    try:
        dt.strptime(race_date, "%Y-%m-%d")
    except Exception:
        raise ValueError("Date must be in 'YYYY-MM-DD' format")

    df = pl.read_excel(ENERGY_XLS_TEMPLATE.format(
        date=dt.strptime(race_date, "%Y-%m-%d").strftime("%Y%m%d")))
    
    # Clean column names
    df.columns = [col.strip().replace(" ", "").replace(
        "\n", "_").replace('.', '') for col in df.columns]

    df = (df.with_columns(pl.col('RunnerNumber').str.to_integer())
          .with_columns(pl.col('SpeedPRO_Energy_Difference').str.to_integer())
          .select(['RaceNumber', 'RunnerNumber', 'HorseName', 'FitnessRatings','SpeedPRO_Energy_Difference']))

    return df


def speedmap(race_num: int) -> str:
    """Fetch SpeedMap as base64 encoded string.

    Args:
        race_date (str): Date in 'YYYY-MM-DD' format.
        race_num (int): Race number.
    
    Returns:
        str: Base64 encoded string of the SpeedMap image.
    """
    r = requests.get(SPEEDMAP_URL_TEMPLATE.format(race_num=race_num))
    if r.status_code != 200:
        raise RuntimeError(f"Request failed: {r.status_code} - {r.text}")
    # Decode with 'utf-8-sig' to strip a possible UTF-8 BOM before JSON parsing
    content = r.content.decode("utf-8-sig")
    return json.loads(content)['en-us']['RaceMap']